/*
 * O'Mix v0.23 : Audio Mixer for Linux (and other Un*x)
 * Copyright (C) 1998 Frank lin PIAT (fpiat@bigfoot.com)
 * 
 * This program is free software, you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of this license, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WANRANTY; without even the implied warranty of
 * MERCHANDABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the free software 
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <xview/generic.h>
#include <xview/xview.h>
#include <xview/frame.h>
#include <xview/panel.h>
#include <xview/openmenu.h>
#include <xview/svrimage.h>
#include <linux/soundcard.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>

#include "omix.h"
#ifdef PICTURE
#include "bitmaps/16x16x1/bitmaps.h"
//#include "bitmaps/24x24x1/bitmaps.h"
#endif
/* Declare Functions... */
void adjust_timer_speed(int value);
Notify_value on_timer();
void update_all();
void quit();
#ifdef PREFS
void ChangeSaveOnExit(Panel_item item, int value, Event *ev);
void customChangeSelected(Panel_item item, int value, Event *ev);
void customChangeOrder(Panel_item item, int value, Event *ev);
void RefreshXViewPrefFrame();
void BuildXViewPrefFrame();
#endif
void RefreshXViewInterface();
void BuildXViewInterface();
int GetMixerInfo();
void menuHandler( Menu item , Menu_item thing, Event *event);
void buttonHandler(Panel_item item, int value);
#ifdef EASYMOVE
void panelHandler( Xv_Window window, Event *event); 
#endif EASYMOVE
//void frameHandler( Xv_Window window, Event *event , Notify_arg arg);
void display_setting(Panel_item item, int value, Event *ev);
void SetLevel( Panel_item, int value, Event *ev );
void SetRecSrc( Panel_item item, int value, Event *ev );
#ifdef RCFILE
void SaveRcFile();
void ReadRcFile();
#endif
void usage();
int parse_args(int argc,char *argv[]);

void quit()
{
xv_destroy_safe(frame);
}
 
void adjust_timer_speed(int value)
{
    if (value > 0) {
        timer.it_value.tv_usec = (value + 20) * 1000;
        timer.it_interval.tv_usec = (value + 20) * 1000;
        notify_set_itimer_func(frame, on_timer, ITIMER_REAL, &timer, NULL);
    } else
        /* turn it off */
        notify_set_itimer_func(frame, NULL, ITIMER_REAL, NULL, NULL);
}

Notify_value on_timer()
{
    update_all();
    return NOTIFY_DONE;
}



void display_setting(Panel_item item, int value, Event *ev)
{
  int temp0,temp1;
//      temp0 = xv_get( item, XV_X);
 //     temp1 = xv_get( item, PANEL_DIRECTION);
      fprintf(stderr," value%d, Event=%d\n",value , ev);
}



#ifdef PREFS
void ChangeSaveOnExit(Panel_item item, int value, Event *ev)
{
   mix_dat.flags = (value) ? mix_dat.flags | TKM_SAVE_ON_EXIT : mix_dat.flags & ~TKM_SAVE_ON_EXIT;
   #ifdef  DEBUG
     printf("SAVE: Now turned %s\n",(value) ? "On" : "Off" );
   #endif
   #ifdef RCFILE
   SaveRcFile();
   #endif
}

void customChangeSelected(Panel_item item, int value, Event *ev)
{
   pref_selected = (xv_get(item, XV_Y) - 5 ) / pref_custom_gui_step;
    #ifdef  DEBUG
   printf("PREF: Selecting item %d (index%d) in custom view\n", pref_selected, deviceCustomTemp[pref_selected] );
   #endif
   xv_set( pref_left, XV_Y, xv_get(item, XV_Y), NULL);
}



void customChangeOrder(Panel_item item, int value, Event *ev)
{
   int this, other,thisD,otherD ,btn;

   this = pref_selected;
   if ( xv_get( item, XV_Y) != xv_get( pref_up_arrow, XV_Y ) ){
	other = this + 1;
	if (deviceCustomTemp[other] == 0xFF) other = this;
   } else {
	other = (this>0) ? this - 1 : this;
   }
   pref_selected = other;

   #ifdef  DEBUG
   printf("PREF: Swaping %d and %d (index%d and %d) in custom view\n", deviceCustomTemp[this], deviceCustomTemp[other],this, other);
   #endif
   if ( this != other) {
	thisD = deviceCustomTemp[this];
	otherD = deviceCustomTemp[other];  
	xv_set(pref_custom_deviceViewed[thisD],	
		XV_SHOW,FALSE,
		NULL);
	xv_set(pref_custom_deviceViewed[otherD],
		XV_Y,5+ this * pref_custom_gui_step,
		NULL);
	xv_set(pref_custom_deviceViewed[thisD],
		XV_Y, 5+ other *pref_custom_gui_step,
		XV_SHOW,TRUE,
		NULL);
	xv_set(pref_custom_deviceIcon[thisD],
		XV_SHOW,FALSE,
		NULL);
	xv_set(pref_custom_deviceIcon[otherD],
		XV_Y,5+ this * pref_custom_gui_step, 
		NULL);
	xv_set(pref_custom_deviceIcon[thisD],  
		XV_Y,5+ other * pref_custom_gui_step,
		XV_SHOW,TRUE,
		NULL);
	xv_set( pref_left,
		XV_Y,5+ other * pref_custom_gui_step,
		NULL);
	deviceCustomTemp[this] = otherD;
	deviceCustomTemp[other] = thisD;
	}

}


void RefreshXViewPrefFrame( /*struct MixerData *mix_dat */ )
{
   int i,count,revCount,ShowInCustom;

   xv_set( prefCustomPanel,
	XV_X,20,
	XV_Y,20,
	NULL);
   xv_set( pref_custom_view_label,
	XV_X,25,
	XV_Y,xv_get(prefCustomPanel, XV_Y) - xv_get(pref_custom_view_label, XV_HEIGHT),
	XV_SHOW, TRUE,
	NULL);

   count = 0;
   ShowInCustom = 1;
   #ifdef DEBUG
      printf("PREF: Refresh");	
   #endif
   for (revCount=0 ; revCount<=20 ; revCount++) {
//      if ( count <= SOUND_MIXER_NRDEVICES ) {
	i = device_index_custom[ revCount ]; 
	#ifdef DEBUG
	  printf(" %d",i );
	#endif
	if (i == 0xFF) { 
	   ShowInCustom = 0; 
	#ifdef DEBUG
	  printf("=HIDE: " );	
	#endif   
	} else if (mix_dat.devmask & (1 << i) ) {
	   xv_set(pref_custom_deviceIcon[i],
		XV_X, 10 + devices_bitmaps_width,
		XV_Y, 5 + pref_custom_gui_step * count,
		NULL);
	   xv_set(pref_custom_deviceViewed[i],
//		PANEL_LABEL_BOXED,TRUE,
//		PANEL_LABEL_INVERTED,TRUE,
//		PANEL_LABEL_BOLD,TRUE ,
		XV_Y, 5 + pref_custom_gui_step * count,
		PANEL_VALUE,  ShowInCustom ,
		XV_X, pref_custom_gui_step * 5 + devices_bitmaps_width 
			- xv_get( pref_custom_deviceViewed[i], XV_WIDTH),
		NULL);
	   deviceCustomTemp[count] = i;
	   count++;
	} // end if
	else {
	   #ifdef DEBUG
           printf("X");
           #endif
	}
//    } //end if
   }// end for
   #ifdef DEBUG
     printf("\n");
   #endif
   pref_selected = 0;
  xv_set(pref_left,
	XV_X,5,
	XV_Y, 5 + pref_custom_gui_step * pref_selected,
	NULL);
   deviceCustomTemp[count] = 0xFF;
 window_fit( prefCustomPanel);
   #ifdef RCFILE
   xv_set(pref_save_on_exit,
	XV_X,xv_get( prefCustomPanel , XV_X),
	XV_Y,xv_get( prefCustomPanel , XV_Y) 
	     + xv_get( prefCustomPanel , XV_HEIGHT) +10,
	PANEL_VALUE,  (mix_dat.flags & TKM_SAVE_ON_EXIT) != 0 ,
	NULL); 
   #endif RCFILE
   xv_set(pref_up_arrow,
	XV_X,xv_get( prefCustomPanel , XV_X) 
		+ xv_get( prefCustomPanel , XV_WIDTH) +20,
	XV_Y, xv_get( prefCustomPanel , XV_Y),
	NULL);
   xv_set(pref_down_arrow,
	XV_X,xv_get( pref_up_arrow ,XV_X),
	XV_Y,xv_get( pref_up_arrow , XV_Y) +10+ devices_bitmaps_height,
	NULL);
   xv_set(pref_ok,
	XV_X,xv_get( pref_up_arrow ,XV_X),
	XV_Y,100 ,//xv_get( pref_down_arrow ,XV_Y) +2* devices_bitmaps_height,
	NULL);
   xv_set(pref_cancel,
	XV_X,xv_get( prefCustomPanel , XV_X) 
	   + xv_get( prefCustomPanel , XV_WIDTH) +20,
	XV_Y,130, 
	NULL);
   #ifdef RCFILE
   xv_set(pref_save,
	XV_X,xv_get( pref_cancel ,XV_X),
	XV_Y,xv_get( pref_save_on_exit ,XV_Y),
	NULL);
   #endif // RCFILE
   xv_set( prefPanel,
	XV_WIDTH, xv_get( pref_cancel, XV_X) 
		+ xv_get( pref_cancel , XV_WIDTH) 
		+  xv_get( prefCustomPanel , XV_X),
	XV_HEIGHT,xv_get( prefCustomPanel , XV_Y)
		#ifdef RCFILE
		+ xv_get( pref_save_on_exit , XV_HEIGHT) 
		+ xv_get( pref_save_on_exit , XV_Y) ,
		#else // RCFILE
		+ xv_get( prefCustomPanel , XV_HEIGHT) 
		+ xv_get( prefCustomPanel , XV_Y),
		#endif // RCFILE
	NULL);
   window_fit( prefFrame);
}



void BuildXViewPrefFrame( /*struct MixerData *mix_dat */ )
{
   int i,count,revCount,ShowInCustom;
   char *devnames[SOUND_MIXER_NRDEVICES] = SOUND_DEVICE_LABELS;

   prefFrame = (Frame)xv_create( frame, FRAME_CMD,
	FRAME_LABEL, "Preference",
	NULL);
   prefPanel =  xv_get( prefFrame , FRAME_CMD_PANEL);
   prefCustomPanel = (Panel)xv_create( prefPanel , PANEL,
	PANEL_LAYOUT, PANEL_HORIZONTAL,
	PANEL_BORDER, TRUE,
	NULL);
    pref_custom_view_label = (Panel)xv_create( prefPanel , PANEL_MESSAGE,
	PANEL_LABEL_STRING, TKM_PREF_CUSTOM_TITLE,
	XV_SHOW, FALSE,
	NULL);

   count = 0;
   ShowInCustom = 1;
   #ifdef DEBUG
      printf("PREF: Creating ");	
   #endif
   for (i=0 ; i<=20 ; i++) {
	if (mix_dat.devmask & (1 << i) ) {
	   #ifdef DEBUG
	     printf(" %d",i );
	   #endif
	   pref_custom_deviceIcon[i] = xv_create(prefCustomPanel, PANEL_MESSAGE,
//		PANEL_LABEL_STRING, devnames[i] ,
		PANEL_LABEL_IMAGE, xv_create( XV_NULL, SERVER_IMAGE,
			XV_WIDTH,		devices_bitmaps_width,
			XV_HEIGHT,		devices_bitmaps_height,
			SERVER_IMAGE_DEPTH,	devices_bitmaps_depth,
			SERVER_IMAGE_BITS,	devices_bitmaps[i],
			NULL),
		PANEL_NOTIFY_PROC, customChangeSelected,
		NULL);
	   pref_custom_deviceViewed[i]= (Panel)xv_create( prefCustomPanel, PANEL_TOGGLE,
		PANEL_FEEDBACK, PANEL_MARKED,
		PANEL_LABEL_STRING, devnames[i] ,
		PANEL_VALUE,  ShowInCustom ,
		NULL);
	   deviceCustomTemp[count] = i;
	   count++;
	} // end if
   }// end for
   #ifdef DEBUG
     printf("\n");
   #endif
   pref_selected = 0;
   pref_left = xv_create(prefCustomPanel, PANEL_MESSAGE,
//	PANEL_LABEL_STRING,">>",
	PANEL_LABEL_IMAGE, xv_create( XV_NULL, SERVER_IMAGE,
		XV_WIDTH,		devices_bitmaps_width,
		XV_HEIGHT,		devices_bitmaps_height,
		SERVER_IMAGE_DEPTH,	devices_bitmaps_depth,
		SERVER_IMAGE_BITS,	move_bitmaps[0],
		NULL),
	PANEL_NOTIFY_PROC, customChangeOrder,
	NULL);
   deviceCustomTemp[count] = 0xFF;
   #ifdef RCFILE
   pref_save_on_exit = (Panel)xv_create( prefPanel, PANEL_TOGGLE,
	PANEL_FEEDBACK, PANEL_MARKED,
	PANEL_LABEL_STRING, TKM_PREF_SAVE_ON_EXIT,
	PANEL_NOTIFY_PROC, ChangeSaveOnExit,
	NULL); 
   #endif // RCFILE
   pref_up_arrow = xv_create(prefPanel, PANEL_MESSAGE,
	PANEL_LABEL_STRING, "Move Up" ,
	PANEL_LABEL_IMAGE, xv_create( XV_NULL, SERVER_IMAGE,
		XV_WIDTH,		devices_bitmaps_width,
		XV_HEIGHT,		devices_bitmaps_height,
		SERVER_IMAGE_DEPTH,	devices_bitmaps_depth,
		SERVER_IMAGE_BITS,	move_bitmaps[1],
		NULL),
	PANEL_NOTIFY_PROC, customChangeOrder,
	NULL);
    pref_down_arrow = xv_create(prefPanel, PANEL_MESSAGE,
	PANEL_LABEL_STRING, "Move Down" ,
	PANEL_LABEL_IMAGE, xv_create( XV_NULL, SERVER_IMAGE,
		XV_WIDTH,		devices_bitmaps_width,
		XV_HEIGHT,		devices_bitmaps_height,
		SERVER_IMAGE_DEPTH,	devices_bitmaps_depth,
		SERVER_IMAGE_BITS,	move_bitmaps[2],
		NULL),
	PANEL_NOTIFY_PROC, customChangeOrder,
	NULL);

   pref_ok = xv_create( prefPanel, PANEL_BUTTON,
		PANEL_LABEL_STRING, TKM_BUTTON_OK ,
		PANEL_NOTIFY_PROC, buttonHandler,
		// TODO use ...CLIENT_DATA, prefFrame,
		NULL);
   pref_cancel = xv_create( prefPanel, PANEL_BUTTON,
		PANEL_LABEL_STRING, TKM_BUTTON_CANCEL ,
		PANEL_NOTIFY_PROC, buttonHandler,
		// TODO use ...CLIENT_DATA, prefFrame,
		NULL);
   #ifdef RCFILE
   pref_save = xv_create( prefPanel, PANEL_BUTTON,
		PANEL_LABEL_STRING, TKM_BUTTON_SAVE_NOW,
		PANEL_NOTIFY_PROC, buttonHandler,
		NULL);
   #endif //RCFILE
   pref_custom_gui_step = ( xv_get( pref_custom_deviceViewed[0] , XV_HEIGHT) >   xv_get( pref_custom_deviceViewed[0] , XV_HEIGHT) ) ? xv_get( pref_custom_deviceViewed[0] , XV_HEIGHT) :   xv_get( pref_custom_deviceViewed[0] , XV_HEIGHT) ;
}
#endif // ifdef PREFS


void menuHandler( Menu item , Menu_item thing, Event *event)
{
  int OutDevice,i;
// TODO replace xv_get(thing, MENU_STRING), with a var for all !
  #ifdef DEBUG
  printf("MENU: \"%s\"\n", xv_get(thing, MENU_STRING) );
  #endif
  #ifdef PREFS
  if (!strcmp((char*)xv_get(thing, MENU_STRING), TKM_MENU_PREFERENCES)) {
	xv_set( prefFrame, XV_SHOW, TRUE, NULL);
  }
  #endif
  if ( (!strcmp((char*)xv_get(thing, MENU_STRING), TKM_MENU_QUIT))
     | (!strcmp((char *)xv_get(thing, MENU_STRING), TKM_BUTTON_MAINMENU)) ) {
	quit();
  }
  if (!strcmp((char *)xv_get(thing, MENU_STRING), TKM_MENU_VIEW_ONE)) {
	mix_dat.flags = (mix_dat.flags & ~TKM_DISPLAY_WIDTH_MASK) | TKM_DISPLAY_WIDTH_ONE;
        RefreshXViewInterface();
  }
  if (!strcmp((char *)xv_get(thing, MENU_STRING), TKM_MENU_VIEW_CUSTOM)) {
        mix_dat.flags = (mix_dat.flags & ~TKM_DISPLAY_WIDTH_MASK) | TKM_DISPLAY_WIDTH_CUSTOM;
        RefreshXViewInterface();
  }
  if (!strcmp((char *)xv_get(thing, MENU_STRING), TKM_MENU_VIEW_LARGE)) {
        mix_dat.flags = (mix_dat.flags & ~TKM_DISPLAY_WIDTH_MASK) | TKM_DISPLAY_WIDTH_LARGE;
        RefreshXViewInterface();
  }
  if (!strcmp((char *)xv_get(thing, MENU_STRING), TKM_MENU_VIEW_ALL)) {
        mix_dat.flags = (mix_dat.flags & ~TKM_DISPLAY_WIDTH_MASK) | TKM_DISPLAY_WIDTH_ALL;
        RefreshXViewInterface();
  }
  if (!strcmp((char *)xv_get(thing, MENU_STRING), TKM_MENU_DISPLAY_LEFT_PANEL)) {
        mix_dat.flags = mix_dat.flags ^ TKM_DISPLAY_LEFT_PANEL;
        RefreshXViewInterface();
  }
  if (!strcmp((char *)xv_get(thing, MENU_STRING), TKM_MENU_TOGGLE_REC)) {
// TODO add a mark in the menu 
// xv_set( viewMenu,  ,OLGX_DIAMOND_MARK,NULL)
        mix_dat.flags = mix_dat.flags ^ TKM_DISPLAY_REC;
        RefreshXViewInterface();
  }
  if (!strcmp((char *)xv_get(thing, MENU_STRING), TKM_MENU_TOGGLE_BAL)) {
        mix_dat.flags = mix_dat.flags ^ TKM_DISPLAY_BAL;
        RefreshXViewInterface();
  }
  if (!strcmp((char *)xv_get(thing, MENU_STRING), TKM_MENU_POLL)) {
	adjust_timer_speed( (mix_dat.flags & TKM_POLL) ? 0 : TKM_POLL_DEFAULT_INTERVAL  );
        mix_dat.flags = mix_dat.flags ^ TKM_POLL;
  }
  if (!strcmp((char *)xv_get(thing, MENU_STRING), TKM_MENU_RECENTER_BALANCE)) {
	OutDevice = device_index_gui[( menuPopX - 10) 
			/ device_step_gui] ;
	#ifdef DEBUG
	printf("USER: Recenter Device:%2d",OutDevice);
	#endif
	if ((1 << OutDevice ) & mix_dat.stereodevs) {
	    i = ( (slider_last_value[OutDevice] & 0x7F)
		> ((slider_last_value[OutDevice] >> 8) & 0x7F) )
		? slider_last_value[OutDevice] & 0x7F
		: (slider_last_value[OutDevice] >> 8) & 0x7F;
	  #ifdef DEBUG
	  printf(" Level  %d\n",i);
	  #endif
	  xv_set( sliderBal[OutDevice]  , PANEL_VALUE, 100, NULL);
	  xv_set( slider[OutDevice], PANEL_VALUE, i  , NULL);
	  i |=(i << 8); 
	  slider_last_value[OutDevice] = i;
	  if (ioctl(mix_dat.mixer_fd,MIXER_WRITE(OutDevice),&i) == -1) {
	     fprintf(stderr,"Error doing IOCTL in SetLevel\n");
	     quit();
	     exit(1);
	  }
	}
	else {
	  #ifdef DEBUG
	    printf(" Can't recenter: device is mono !\n");
	  #endif
	}
  }
  if ( (!strcmp((char *)xv_get(thing, MENU_STRING), TKM_MENU_VIEW_NEXT))
     | (!strcmp((char *)xv_get(thing, MENU_STRING), TKM_BUTTON_VIEWMENU)) ) { 
	if ((mix_dat.flags & TKM_DISPLAY_WIDTH_ALL) ){
 	       mix_dat.flags = (mix_dat.flags & ~TKM_DISPLAY_WIDTH_MASK) | TKM_DISPLAY_WIDTH_ONE ;
	} else {
        	mix_dat.flags = (mix_dat.flags & ~TKM_DISPLAY_WIDTH_MASK) | ( (mix_dat.flags & TKM_DISPLAY_WIDTH_MASK) << 1 );
	}
        RefreshXViewInterface();
  }

}


void buttonHandler(Panel_item item, int value)
{
   int i,count;
   #ifdef DEBUG
   printf("BTTN: \"%s\"\n", xv_get(item, PANEL_LABEL_STRING) );
   #endif
   #ifdef PREFS 
   if (!strcmp((char *)xv_get(item, PANEL_LABEL_STRING), TKM_BUTTON_CANCEL)) {
	if ((int)xv_get(prefFrame, FRAME_CMD_PIN_STATE) == FRAME_CMD_PIN_OUT)
	   xv_set( prefFrame, XV_SHOW, FALSE, NULL);
	RefreshXViewPrefFrame();
	RefreshXViewPrefFrame();
   }
   if (!strcmp((char *)xv_get(item, PANEL_LABEL_STRING), TKM_BUTTON_OK)) {
	count= 0;
	i = 0;
	while ( (i <= 21 ) & (deviceCustomTemp[i]!=0xFF) ) { 
	  if (mix_dat.devmask & (1 << deviceCustomTemp[i])) {
	    if (deviceCustomTemp[i] < SOUND_MIXER_NRDEVICES) { 
		if (xv_get( pref_custom_deviceViewed[deviceCustomTemp[i]], PANEL_VALUE) ) {
		   device_index_custom[count] = deviceCustomTemp[i];
		   count++;
		} 
	    }
	  }
	i++;
	} // end while
	device_index_custom[count]=0xFF;
	mix_dat.flags = (mix_dat.flags & ~TKM_DISPLAY_WIDTH_MASK) | TKM_DISPLAY_WIDTH_CUSTOM;
	RefreshXViewInterface();
   }
  #ifdef RCFILE 
  if (!strcmp((char *)xv_get(item, PANEL_LABEL_STRING), TKM_BUTTON_SAVE_NOW)){
	SaveRcFile();
  }
  #endif // ifdef RCFILE
  #endif // ifdef PREFS
  if (!strcmp((char *)xv_get(item, PANEL_LABEL_STRING), TKM_BUTTON_VIEW_ONE)){
	mix_dat.flags = (mix_dat.flags & ~TKM_DISPLAY_WIDTH_MASK) | TKM_DISPLAY_WIDTH_ONE;
	RefreshXViewInterface( );
   }
   if (!strcmp((char *)xv_get(item,PANEL_LABEL_STRING),TKM_BUTTON_VIEW_CUSTOM)){
        mix_dat.flags = (mix_dat.flags & ~TKM_DISPLAY_WIDTH_MASK) | TKM_DISPLAY_WIDTH_CUSTOM;
        RefreshXViewInterface( );
   }
   if (!strcmp((char *)xv_get(item, PANEL_LABEL_STRING), TKM_BUTTON_VIEW_ALL)){
        mix_dat.flags = (mix_dat.flags & ~TKM_DISPLAY_WIDTH_MASK) | TKM_DISPLAY_WIDTH_ALL;
        RefreshXViewInterface( );
   }
}


void SetLevel( Panel_item item, int value, Event *ev )
{
  int OutValue,OutDevice,Bal,Level;

  OutDevice = device_index_gui[(xv_get( item, XV_X) - 10) / device_step_gui]; 
  #ifdef DEBUG
  printf("USER: Device:%2d Level  %d\n",OutDevice,value);
  #endif
  if (xv_get( item, PANEL_DIRECTION) == PANEL_HORIZONTAL ) {
    Level = xv_get( slider[OutDevice] , PANEL_VALUE);
    Bal = value;
  } else {
     Level = value;
     Bal = xv_get( sliderBal[OutDevice] , PANEL_VALUE);
  }
  if ((1 << OutDevice ) & mix_dat.stereodevs) {
     // Device is stereo
//     fprintf(stderr,"USER: newvalue =%3d sliderBal[i]=%3d\n",Level ,Bal);
     if (Bal==100) {
	 OutValue = (Level << 8) | (Level);
     } else 
     if ( Bal > 100) {
	// Balance is on Right Side
	OutValue = (( Level  ) << 8) | ( Level - Level*(Bal-100)/100 );
     } else {
	// Balance is on Left Side
	OutValue = (( Level*(Bal)/100  ) << 8) | (Level);
     }
//     fprintf(stderr,"AUTO: left:%3d right%3d (0x%2x)\n",(OutValue >> 8) & 0xFF,OutValue &  +0xFF,OutValue);
  } else {
     // Device is Mono  
     OutValue = (( Level  ) << 8) | Level;
  }
  slider_last_value[OutDevice] = OutValue  ;
  if ( ioctl(mix_dat.mixer_fd, MIXER_WRITE(OutDevice), &OutValue)  == -1) {
     fprintf(stderr,"Error doing IOCTL in SetLevel\n");
     quit();
     exit(1);
  }
}



void SetRecSrc( Panel_item item, int value, Event *ev )
{
   int OutDevice;
   OutDevice = device_index_gui[(xv_get( item, XV_X) - 10) / device_step_gui]; 
   #ifdef DEBUG
   printf("USER: Device:%2d record %s\n",OutDevice, (value) ? "On" : "Off");
   #endif
   OLD_mix_dat_recsrc ^= (1 << OutDevice);
   mix_dat.recsrc =((mix_dat.recsrc) & ~(1 << OutDevice) | (value << OutDevice) );
   if (ioctl(mix_dat.mixer_fd, SOUND_MIXER_WRITE_RECSRC, &(mix_dat.recsrc)) == -1) {
      fprintf(stderr,"Error doing IOCTL in SetRecSrc\n");
      quit();
      exit(1);
   }
}



int GetMixerInfo( /*struct MixerData *mix_dat*/ )
{
   /* A hack for the time being */
   /* HTF do you find out what mixer /dev/mixer points to? */
   #ifdef PAS
      mix_dat.card=SNDCARD_PAS;
   #endif

   /* Open the Mixer */
   #ifdef DEBUG
     printf("INIT: Opening Mixer %s\n",mixer_device);
   #endif
   if ((mix_dat.mixer_fd = open(mixer_device, O_RDWR)) <0) {
      fprintf(stderr, "Error opening %s\n",mixer_device);
      return(0);
   }

   /* Read bit mask of available devices */
   if (ioctl(mix_dat.mixer_fd, SOUND_MIXER_READ_DEVMASK, &mix_dat.devmask) == -1) {
     fprintf(stderr,"Error obtaining DEVMASK\n");
     return(0);
   }
   #ifdef DEBUG
      printf("INIT: DevMask:%x\n",mix_dat.devmask);
   #endif
   // Read bit mask of recording devices 
   if (ioctl(mix_dat.mixer_fd, SOUND_MIXER_READ_RECMASK, &mix_dat.recmask) == -1) {
      fprintf(stderr,"Error obtaining RECMASK\n");
      return(0);
   }
   #ifdef DEBUG
      printf("INIT: RecMask:%x\n",mix_dat.recmask);
   #endif

   /* Read bit mask of stereo devices */
   if (ioctl(mix_dat.mixer_fd, SOUND_MIXER_READ_STEREODEVS, &mix_dat.stereodevs) == -1) {
      fprintf(stderr,"Error obtaining STEREODEVS\n");
      return(0);
   }
   #ifdef DEBUG
        printf("INIT: StereoMask:%x\n",mix_dat.stereodevs);
   #endif
   return(1);
}



/*
void frameHandler( Xv_Window window, Event *event , Notify_arg arg)
{
// printf("window %x,Event %x,Action %x, arg %x\n",window,  event,event_action(event), arg);
   if( event_id( event ) == WIN_RESIZE) {
   #ifdef DEBUG
	printf(" The window was resized\n");
	// TODO Handle this event !
   #endif
   }
}
*/


void panelHandler( Xv_Window window, Event *event) {
#ifdef EASYMOVE
   if ((event_action(event)== MS_LEFT)|(event_action(event)== ACTION_SELECT)) {
      //printf("MOUSE:");
      if (!event_is_down(event) ) {
	//printf(" button Released :");
	if (DragMsWasUp) {
	   DragMsWasUp = FALSE;
	   //printf(" DROP at   (%d,%d)\n",event_x(event),event_y(event) );
	} else {
	   //printf(" Ignore, we were in SetLevel\n");
	   DragCanStart = TRUE;
	}
      }
    // else printf(" button Pressed  :");
   }	

   if (event_id(event) == LOC_DRAG) {
	if (!DragMsWasUp) {
	   if ((event_y(event) < (int)xv_get(slider[0],XV_Y)) & DragCanStart) {
		DragMsWasUp = TRUE;
		DragXoffset = event_x(event);
		DragYoffset = event_y(event);
		//printf(" DRAG from (%d,%d)\n",event_x(event),event_y(event) );
	   } 
	 else 
	   {
		//printf(" in SetLevel!\n");
		DragCanStart = FALSE;
	   }
	} else { 
	//printf(" moved to (%d,%d)\n", event_x(event), event_y(event) );
	xv_set( frame,
		XV_X,xv_get(frame,XV_X) - DragXoffset + event_x(event),
		XV_Y,xv_get(frame,XV_Y) - DragYoffset + event_y(event),
		NULL);
	}
   }
   #endif // EASYMOVE
   if (event_action(event) == ACTION_MENU && event_is_down(event) ) {
	Menu menu = (Menu)xv_get(window , WIN_CLIENT_DATA);
	menu_show( menu, window, event, NULL);
	menuPopX = event_x(event);
	// TODO hide recenter if device is mono !
	// TODO set the panel's title according to the current device
   }
}


void RefreshXViewInterface()
{
   int i,count,first,last;
   int y_rec, y_slider,y_bal,y_label;

   if (mix_dat.flags & TKM_DISPLAY_LEFT_PANEL) {
 	xv_set( leftPanel,	XV_SHOW, TRUE,NULL);
	xv_set( viewMenu,	XV_SHOW, TRUE,NULL);
	xv_set( mainMenu,	XV_SHOW, TRUE,NULL);
	xv_set( rightPanel,	XV_X,
		xv_get(leftPanel,XV_WIDTH) + xv_get(leftPanel, XV_X) * 2 , 
		NULL);	
  } else {
 	xv_set( leftPanel,	XV_SHOW, FALSE,NULL);
	xv_set( viewMenu,	XV_SHOW, FALSE,NULL);
	xv_set( mainMenu,	XV_SHOW, FALSE,NULL);
 	xv_set( rightPanel,	//XV_SHOW, FALSE,	
				XV_X,xv_get(leftPanel, XV_X),
				//XV_SHOW,TRUE,
				NULL );
   }
   for(i=0;i<=SOUND_MIXER_NRDEVICES;i++) {
      if((1 << i) & mix_dat.devmask) {
	  xv_set( slider[i]     ,XV_SHOW, FALSE,NULL);
          xv_set( sliderRec[i]  ,XV_SHOW, FALSE,NULL);
          xv_set( sliderBal[i]  ,XV_SHOW, FALSE,NULL);
          xv_set( sliderLabel[i],XV_SHOW, FALSE,NULL);
      }
   }

   y_label = 5;
   y_rec = 10  + xv_get( sliderLabel[0] , XV_HEIGHT) 
   #ifndef PICTURE
   * 2
   #endif
   ;
   if ( mix_dat.flags & TKM_DISPLAY_REC ) { 
	y_slider = y_rec + 0 + xv_get(sliderRec[0], XV_HEIGHT);
   } else { y_slider = y_rec;}
	y_bal  = y_slider + 0 + xv_get(slider[0], XV_HEIGHT) ;

   // TODO , standardize the use of count, first,last
   // device_index_gui[ count] = 0xFF.... to optimise and improve readability
   // HERE, we create a table with the devices to be displayed
   switch (mix_dat.flags & TKM_DISPLAY_WIDTH_MASK) {
	case TKM_DISPLAY_WIDTH_ONE:
	   #ifdef DEBUG
	      printf("DISP: One ");
	   #endif
	   count =1;
	   device_index_gui[0] = SOUND_MIXER_VOLUME;
	   device_index_gui[1] = 0xFF;
	   break;
	case TKM_DISPLAY_WIDTH_CUSTOM:
	   #ifdef DEBUG
	      printf("DISP: Custom ");
	   #endif
	   count =0;
	   i =0;
	   while (device_index_custom[count]!=0xFF & i<20) {
	     if ((1 << i) & mix_dat.devmask ) {
		device_index_gui[count] = device_index_custom[i];
		count++;
	     }
	     i++;
	   }
	   device_index_gui[count] = 0xFF;
           break;
        case TKM_DISPLAY_WIDTH_LARGE:
	    #ifdef DEBUG
	      printf("DISP: Large ");
	    #endif
	    count = -1;
            first = 0 ; last = SOUND_MIXER_NRDEVICES;
            for(i= first ; i <= last ;i++) {
	        if ( (i != SOUND_MIXER_IGAIN) & (i != SOUND_MIXER_OGAIN) )
		  if ((1 << i) & mix_dat.devmask ) 
                     device_index_gui[++count] = i;
             } // end FOR
            device_index_gui[count+1] = 0xFF;
            break;
        case TKM_DISPLAY_WIDTH_ALL:
	    #ifdef DEBUG
	      printf("DISP: All ");
	    #endif
	    count = -1;
	    first = 0 ; last = SOUND_MIXER_NRDEVICES;
	    for(i= first ; i <= last ;i++) {
	       if((1 << i) & mix_dat.devmask) {
	          device_index_gui[++count] = i;
	       }
	    } // end FOR
            device_index_gui[count+1] = 0xFF; 
            break;
	default:
	    #ifdef DEBUG
	      printf("DISP: Unknown Menu Option (%d)\n",mix_dat.flags & TKM_DISPLAY_WIDTH_MASK);
	    #else
	      fprintf(stderr,"Unknown Menu Option (%d)\n",mix_dat.flags & TKM_DISPLAY_WIDTH_MASK);
	    #endif
	   device_index_gui[0] = 0xFF;
	   break;
   }// END SWITCH

   // TODO  we could use count to know how many devices to display
   // HERE, we display the chanels included in the table define above.
   count =0;
   while (device_index_gui[count]!=0xFF & count<20) {
      i = device_index_gui[count] ;
      #ifdef DEBUG
      printf("%d ",i);
      #endif
      if((1 << i) & mix_dat.devmask) {
 	 xv_set( sliderLabel[i],
                XV_Y, y_label
		#ifndef PICTURE
		 + (count & 1) * 10
		#endif
		,XV_X,device_step_gui * count +10,
                XV_SHOW, TRUE,
                NULL);
         if ( mix_dat.flags & TKM_DISPLAY_REC) if ((1 << i) & mix_dat.recmask) {
           xv_set( sliderRec[i],
		XV_Y, y_rec,
                XV_X,device_step_gui * count +10,
                XV_SHOW, TRUE,
                NULL);
	 }
	 xv_set( slider[i],
                XV_Y, y_slider,
                XV_X,device_step_gui * count +10,
                XV_SHOW, TRUE,
                NULL);
	 if ( mix_dat.flags & TKM_DISPLAY_BAL)
	  if ( mix_dat.stereodevs & (1 << i) )
            xv_set( sliderBal[i],
		XV_Y, y_bal,
                XV_X,device_step_gui * count +10,
                XV_SHOW, TRUE,
                NULL);
	 count++;
      } else {
	 count++;
	 #ifdef DEBUG
         printf(":SKIP ");
	 #else
         fprintf(stderr,"This (%d) device doesn't exist... SKIPed\n",i);
	 #endif
      }
   } // END WHILE
   #ifdef DEBUG
   printf("...done\n");
   #endif
   xv_set( leftPanel,
	XV_SHOW,FALSE,
	XV_HEIGHT,( mix_dat.flags & TKM_DISPLAY_BAL)
		? y_bal + xv_get(sliderBal[0],XV_HEIGHT) 
		: y_bal ,
	XV_SHOW, (mix_dat.flags & TKM_DISPLAY_LEFT_PANEL)
		? TRUE
		: FALSE ,
	NULL);
   
   xv_set( rightPanel,
        XV_SHOW, FALSE,
	XV_HEIGHT, xv_get(leftPanel,XV_HEIGHT),
        XV_WIDTH, device_step_gui * count + 10,
	PANEL_BORDER,((mix_dat.flags & TKM_DISPLAY_LEFT_PANEL)!=0)
		? TRUE 
		: FALSE,
       NULL);

   xv_set( frame,
	XV_WIDTH,xv_get(rightPanel, XV_X) 
		+ xv_get(rightPanel, XV_WIDTH) + xv_get(leftPanel, XV_X),
	XV_HEIGHT, xv_get(leftPanel, XV_HEIGHT)  
		+ xv_get(leftPanel, XV_Y) + xv_get(leftPanel, XV_X),
        NULL);
   // if we show this panel before resizing the frame,
   // the panel doesn't get resized (properly)
   xv_set( rightPanel,
        XV_SHOW, TRUE,
	NULL);
}




void BuildXViewInterface()
{
   register int i,j;
   int lablen,start,end,gui_count;
   Panel_item btnView,btnSys;    
   char *devnames[SOUND_MIXER_NRDEVICES] = SOUND_DEVICE_LABELS;


   viewMenu = (Menu)xv_create( XV_NULL,  MENU,
	MENU_STRINGS,
                TKM_MENU_VIEW_NEXT,TKM_MENU_VIEW_ONE,TKM_MENU_VIEW_CUSTOM,
		TKM_MENU_VIEW_LARGE,TKM_MENU_VIEW_ALL,"",
                TKM_MENU_TOGGLE_REC,TKM_MENU_TOGGLE_BAL,
		TKM_MENU_DISPLAY_LEFT_PANEL,
		NULL,
	MENU_NOTIFY_PROC,menuHandler,
	NULL);

   mainMenu = (Menu)xv_create( XV_NULL,  MENU,
        MENU_STRINGS,
		TKM_MENU_QUIT,TKM_MENU_POLL,
		#ifdef PREFS
		TKM_MENU_PREFERENCES,
		#endif
		"*Set Gain",TKM_MENU_ABOUT,
                NULL,
        MENU_NOTIFY_PROC,menuHandler,
        NULL);

   fullMenu = (Menu)xv_create( XV_NULL, MENU,
	MENU_NOTIFY_PROC,menuHandler,
	MENU_ITEM,
		MENU_STRING,"View",
		MENU_PULLRIGHT, viewMenu,NULL,
	MENU_ITEM,
		MENU_STRING,"Sys",
		MENU_PULLRIGHT, mainMenu,NULL,
	MENU_STRINGS,
		TKM_MENU_RECENTER_BALANCE,
//TODO		"/Zoom this device","Enable as recsrc",
		NULL,
	NULL);

   frame = (Frame)xv_create(XV_NULL, FRAME,
	FRAME_LABEL, "O'Mix",
//	FRAME_SHOW_RESIZE_CORNER, FALSE,
//	XV_X,50, // it seems to work under openWindows only not twm, fwm...
//	XV_Y,50,
/*	if, at some stage i want to handle user "window resize".
	WIN_CONSUME_EVENTS,
		WIN_NO_EVENTS,
		NULL,
	WIN_EVENT_PROC, frameHandler,*/
	NULL);

   leftPanel = (Panel)xv_create(frame, PANEL,
	PANEL_LAYOUT, PANEL_VERTICAL,
	/* If we do this, menu won't pop from View and Sys button.
   	WIN_CONSUME_EVENTS,WIN_MOUSE_BUTTONS,NULL,
	WIN_EVENT_PROC, panelHandler,
	WIN_CLIENT_DATA, viewMenu,
	*/
	NULL);

   btnView = xv_create(leftPanel, PANEL_BUTTON,
        PANEL_LABEL_STRING, TKM_BUTTON_VIEWMENU,
	PANEL_ITEM_MENU,viewMenu,
        NULL);

   btnSys = xv_create(leftPanel, PANEL_BUTTON,
        PANEL_LABEL_STRING, TKM_BUTTON_MAINMENU,
        PANEL_ITEM_MENU,mainMenu,
	NULL);

   #ifdef BUTTONS
   (void) xv_create(leftPanel, PANEL_BUTTON,
        PANEL_LABEL_STRING, TKM_BUTTON_VIEW_ONE,
        PANEL_NOTIFY_PROC, buttonHandler,
        NULL);
   (void) xv_create(leftPanel, PANEL_BUTTON,
        PANEL_LABEL_STRING, TKM_BUTTON_VIEW_CUSTOM,
        PANEL_NOTIFY_PROC, buttonHandler,
        NULL);
   (void) xv_create(leftPanel, PANEL_BUTTON,
        PANEL_LABEL_STRING, TKM_BUTTON_VIEW_ALL,
        PANEL_NOTIFY_PROC, buttonHandler,
        NULL);
   #endif
//   window_fit( leftPanel );
   i = xv_get( btnView , XV_WIDTH);
   j = xv_get( btnSys  , XV_WIDTH);
   if (j>i) {
      i=j;
//      xv_set( btnView , XV_WIDTH , i * 2, NULL  );
   } else {
//      xv_set( btnSys  , XV_WIDTH , i * 2, NULL );
   }
   xv_set(leftPanel , XV_WIDTH , i + 8, NULL );

   rightPanel = (Panel)xv_create(frame, PANEL,
	PANEL_LAYOUT, PANEL_HORIZONTAL,
	PANEL_BORDER, TRUE,
	WIN_CONSUME_EVENTS,
		WIN_MOUSE_BUTTONS,
		#ifdef EASYMOVE
		  LOC_DRAG,
		#endif
		NULL,
	WIN_EVENT_PROC, panelHandler,
	WIN_CLIENT_DATA, fullMenu,
	XV_SHOW,FALSE,
	NULL);	
    gui_count = -1;

    for(i=0 ; i <= SOUND_MIXER_NRDEVICES ;i++) {
      if((1 << i) & mix_dat.devmask) {
        #ifdef DEBUG
        printf("INIT: Create:%2d %s\n",i,devnames[i]);
        #endif
        device_index_gui[++gui_count] = i;
	sliderRec[i] = (Panel) xv_create( rightPanel, PANEL_TOGGLE,
		PANEL_NOTIFY_PROC, SetRecSrc,
		PANEL_FEEDBACK, PANEL_MARKED,
		NULL);
	slider[i] = xv_create (rightPanel, PANEL_SLIDER,
		PANEL_MIN_VALUE, 0,
		PANEL_MAX_VALUE, 100,
		PANEL_NOTIFY_PROC, SetLevel ,
		PANEL_NOTIFY_LEVEL, PANEL_ALL,
		PANEL_SHOW_RANGE, FALSE,
		PANEL_SHOW_VALUE, FALSE,
		PANEL_SLIDER_WIDTH,100,
		PANEL_DIRECTION,PANEL_VERTICAL,
		NULL);
	sliderBal[i] = xv_create (rightPanel, PANEL_SLIDER,
                PANEL_MAX_VALUE, 200,
                PANEL_NOTIFY_PROC,  SetLevel ,
                PANEL_NOTIFY_LEVEL, PANEL_ALL,
                PANEL_SHOW_RANGE, FALSE,
                PANEL_SHOW_VALUE, FALSE,
                PANEL_SLIDER_WIDTH,15,
                PANEL_DIRECTION,PANEL_HORIZONTAL,
                NULL);

	sliderLabel[i] = xv_create(rightPanel, PANEL_MESSAGE,
		#ifndef PICTURE
        	PANEL_LABEL_STRING, devnames[i] ,
		#else
		PANEL_LABEL_IMAGE, xv_create( XV_NULL, SERVER_IMAGE,
				XV_WIDTH,		devices_bitmaps_width,
				XV_HEIGHT,		devices_bitmaps_height,
				SERVER_IMAGE_DEPTH,	devices_bitmaps_depth,
				SERVER_IMAGE_BITS,	devices_bitmaps[i],
				NULL),
//		PANEL_NOTIFY_PROC, display_setting,
		#endif // !PICTURE
		NULL);
     } // endif
   }  // endfor
   // depending on scale, device_step_gui = 26, 30, 35, 45
   device_step_gui = xv_get( slider[0] , XV_WIDTH)*5/3;
   if(device_step_gui<xv_get( slider[0] , XV_WIDTH) )
	device_step_gui = xv_get( slider[0] , XV_WIDTH);
   #ifdef DEBUG
     printf("INIT: space between 2 sliders=%d\n",device_step_gui);
   #endif
}







void update_all()
{
    int i,temp,L,R;
    char *devnames[SOUND_MIXER_NRDEVICES] = SOUND_DEVICE_LABELS;

// fprintf(stderr,"frame XV_X=%d\n", xv_get( frame, XV_X ) );

    /* Read bit mask of _ALL_ recording sources - at once*/
    if (ioctl(mix_dat.mixer_fd, SOUND_MIXER_READ_RECSRC,&mix_dat.recsrc) == -1)
    {
       fprintf(stderr,"Error obtaining RECSRC\n");
       quit(); 
       exit(1);
    }
      /* for each _existing_ devices */
      for(i=0;i<=SOUND_MIXER_NRDEVICES;i++) {
         if((1 << i) & mix_dat.devmask) {
 	   /* Set the green indicator if the device is selected */
           /* for recording */
           // Record is enabled for this device ?
 	   if ( OLD_mix_dat_recsrc != mix_dat.recsrc ) { 
              #ifdef DEBUG
              printf("READ: Device:%2d recsrc %s",i, (((1 << i) & mix_dat.recsrc)!=0) ? "on " : "off" );
	      #endif
              if ((1 << i) & mix_dat.recsrc) {
	         xv_set(sliderRec[i], PANEL_VALUE, 1,  NULL);
              } else {
	         xv_set(sliderRec[i], PANEL_VALUE, 0, NULL);
              }
           }

          /* Read the value of each channel of the mixer */
          if(ioctl(mix_dat.mixer_fd,MIXER_READ(i),&temp) == -1) {
             fprintf(stderr,"Unable to read status of device %s\n",i);
	     quit();
	     exit(1); 
          }
	  if ( slider_last_value[i] != temp )
          {
             #ifdef DEBUG
//             printf("READ: Device:%2d Level  %d (0x%4X)\n",i, temp & 0x7F , temp);
               printf(" Level %3d (0x%4X)\n", temp & 0x7F , temp);
             #endif

             if ((1 << i) & mix_dat.stereodevs) {
		L = temp & 0x7F;
		R = (temp >> 8)  & 0x7F;
		if (L==R) {
		   xv_set( slider[i] , PANEL_VALUE , L , NULL);
                   xv_set( sliderBal[i] , PANEL_VALUE , 100 , NULL);
		} else
		if (L > R){
		   xv_set( slider[i] , PANEL_VALUE , L , NULL);
		   xv_set( sliderBal[i], PANEL_VALUE, 100- (L-R)*100/L, NULL);
		} else {
                   xv_set( slider[i] , PANEL_VALUE , R , NULL);
		   xv_set( sliderBal[i], PANEL_VALUE, 100+ (R-L)*100/R, NULL);
		}
	     } else {
		/* It's a mono device */
		xv_set( slider[i] , PANEL_VALUE ,  (temp & 0x7F)  , NULL);
	     }
	     slider_last_value[i] = temp;
          } // END IF 
       } /* end if */
    } /* end for */

   /* Changes detected, lets update it ! */
   if ( OLD_mix_dat_recsrc != mix_dat.recsrc )
      OLD_mix_dat_recsrc = mix_dat.recsrc;
}

#ifdef RCFILE
void SaveRcFile()
{
   char		*rcfile_name,*home;
   FILE		*rcfile;
   struct stat  rcfile_stat;
   char		*linePTR,line[255]; // also change below
   char		*argv[32]; // also change below
   int 		i,argc,end;

   if ((home = getenv("HOME")) == NULL) {
	fprintf(stderr, "The envirenemt variable $HOME is not set!\n");
	exit(1);
   }
   rcfile_name = malloc(strlen(home) + sizeof(TKM_RCFILE));
   if (rcfile_name == NULL) {
	perror("SaveRcFile()");
	exit(1);
   }
   strcpy(rcfile_name , home);
   strcat(rcfile_name , TKM_RCFILE);
   #ifdef DEBUG
   fprintf( stderr, "file: Opening %s (to save)\n", rcfile_name );
   #endif
   rcfile = fopen( rcfile_name, "w");
   /* obtain a file handler */
   if ( rcfile == NULL) {
	perror(rcfile_name); //Can't find the file... no problem, forget it!
	return;
//	exit(1);
   } else {
	/* get the file "stat" */
	if ( fstat( fileno( rcfile ) , &rcfile_stat ) < 0) {
	   perror(rcfile_name);
	   return;
//   	   exit(1);
	}
	if( ! S_ISREG(rcfile_stat.st_mode)) {
	   errno = EISDIR;
	   perror( rcfile_name );
	   fclose( rcfile );
	   return;
//   	   exit(1);
	}
   fprintf(rcfile,"%s ", mixer_device );
   fprintf(rcfile,"%cs %cp %cr %cb %cm",\
	((mix_dat.flags & TKM_SAVE_ON_EXIT) == 0) 	? 45 : 43,\
	((mix_dat.flags & TKM_POLL) == 0) 		? 45 : 43,\
	((mix_dat.flags & TKM_DISPLAY_REC) == 0) 	? 45 : 43,\
	((mix_dat.flags & TKM_DISPLAY_BAL) == 0) 	? 45 : 43,\
	((mix_dat.flags & TKM_DISPLAY_LEFT_PANEL) == 0) ? 45 : 43);
 	end = 0xFF;  
 	if ((mix_dat.flags & TKM_DISPLAY_WIDTH_ONE) != 0)  end = 49;  /* 1 */ 
 	if ((mix_dat.flags & TKM_DISPLAY_WIDTH_CUSTOM)!=0) end = 99;  /* c */
 	if ((mix_dat.flags & TKM_DISPLAY_WIDTH_LARGE) !=0) end = 108; /* l */
 	if ((mix_dat.flags & TKM_DISPLAY_WIDTH_ALL) != 0)  end = 100; /* d */
  	if (end != 0xFF) fprintf(rcfile," -v%c",end);
	end =0;
	fprintf( rcfile," --custom=");
	while ((device_index_custom[end] !=0xFF) & (end<21)) {
		fprintf(rcfile,"%d,",device_index_custom[end]);
		end++;
	}
  fclose(rcfile);
   }
}



void ReadRcFile()
{
   char		*rcfile_name,*home;
   FILE		*rcfile;
   struct stat  rcfile_stat;
   char		*linePTR,line[255]; // also change below
   char		*argv[32]; // also change below
   int 		i,argc,end;

   if ((home = getenv("HOME")) == NULL) {
	fprintf(stderr, "The envirenemt variable $HOME is not set!\n");
	exit(1);
   }
   rcfile_name = malloc(strlen(home) + sizeof(TKM_RCFILE));
   if (rcfile_name == NULL) {
	perror("ReadRcFile()");
	exit(1);
   }
   strcpy(rcfile_name , home);
   strcat(rcfile_name , TKM_RCFILE);
   #ifdef DEBUG
   fprintf( stderr, "file: Opening %s\n", rcfile_name );
   #endif
   rcfile = fopen( rcfile_name, "r");
   /* obtain a file handler */
   if ( rcfile == NULL) {
//	perror(rcfile_name); //Can't find the file... no problem, forget it!
	return;
//	exit(1);
 	#ifdef DEBUG
	   perror(rcfile_name);
	#endif
  } else {
	/* get the file "stat" */
	if ( fstat( fileno( rcfile ) , &rcfile_stat ) < 0) {
	   perror(rcfile_name);
	   return;
//   	   exit(1);
	}
	if( ! S_ISREG(rcfile_stat.st_mode)) {
	   errno = EISDIR;
	   perror( rcfile_name );
	   fclose( rcfile );
	   return;
//   	   exit(1);
	}
   do { 
     linePTR = fgets( line, 255, rcfile);
     if (linePTR!=0) {
	#ifdef DEBUG
	printf ("file: processing %s\n", line  );
	#endif
	argc = 2;
	argv[0] = &line[0]; // fake command line
	argv[1] = &line[0];
	end = strlen( line );
	for ( i = 0; i <= end ; i++) {
	   if( (line[i] == 32) | (line[i] == 9) | (line[i] == 10) ) {
		line[i] = 0;
		argv[argc++] = &line[i]+1;
		
	   }
	} // end for
/*	#ifdef DEBUG
	  printf ("file: parse_args \"%s", line  );
	  for ( i = 0 ; i < argc ; i++ ) printf(" %s",argv[i]);
	  printf( "\"\n" );
	#endif
*/
	parse_args(argc, argv);
     }
   } while (linePTR !=0 );
   fclose( rcfile);
   } // end if  fileOpened
} // end func
#endif RCFILE

void usage()
{
        fprintf (stderr,"Usage: omix [mixer] [options]\n");
        fprintf (stderr,"Options:\n");
	fprintf (stderr,"Where mixer is a valid mixer (/dev/mix)\n");
        fprintf (stderr,"  -h --help   Display This\n");
	fprintf (stderr,"  -help       Display _XVIEW_ generic command-line option");
        fprintf (stderr,"  +|-p        Don't poll mixer device\n");
        fprintf (stderr,"  -v1         View Volume Only\n");
        fprintf (stderr,"  -vc         View Custom\n");
        fprintf (stderr,"  -vl         View Large\n");
        fprintf (stderr,"  -vd         View All(Debug)\n");
        fprintf (stderr,"  +|-r        View record source\n");
        fprintf (stderr,"  +|-b        View balance control\n");
        fprintf (stderr,"  +|-m        View left panel\n");
#ifdef RCFILE
	fprintf (stderr,"  +|-s        Save on Exit\n");
#endif
/*
#ifdef PAS
        fprintf (stderr,"  -f          Toggle Stereo Field controls (PAS only)\n");
        fprintf (stderr,"  -l          Toggle Loudness controls (PAS only)\n");
        fprintf (stderr,"  -m          Toggle Mute controls (PAS only)\n");
#endif
*/
}


int parse_args(int argc,char *argv[])
{
   int i, bit_mask, x,len,device,pos;
   #ifdef DEBUG
   printf("PARSE: commandline:[%d] ",argc);
   for (i=0 ; i<argc ; i++)
	printf(" %s",argv[i] );
   printf("\n");
   #endif
   for (i=1;i<argc;i++) {
	if ( argv[i][0] == 47) {			/*  "/dev/"  */
	   strcpy(mixer_device, argv[i]);
	   #ifdef DEBUG
           printf("PARSE: new mixer: %s\n",mixer_device);
           #endif
	   continue;
	}	
        if (!strcmp(argv[i], "-h") | !strcmp(argv[i] ,"--help") ) {
           /* Usage info */
           usage();
           exit(1);
	}
	if ((strlen(argv[i])==2) & ( argv[i][0]==43 | argv[i][0]==45) ) {
	   switch ( argv[i][1] ) {
		case 112 /*p*/:
			bit_mask = TKM_POLL;
			#ifdef DEBUG
			   printf("PARSE: POLL");
			#endif
			break;
		case 98  /*b*/:
			bit_mask = TKM_DISPLAY_BAL;
			#ifdef DEBUG
			   printf("PARSE: BAL ");
			#endif
			break;
		case 114 /*r*/:
			bit_mask = TKM_DISPLAY_REC;
			#ifdef DEBUG
			   printf("PARSE: REC ");
			#endif
			break;
		case 109 /*m*/:
			bit_mask = TKM_DISPLAY_LEFT_PANEL;
			#ifdef DEBUG
			   printf("PARSE: MENU");
			#endif
			break;
		#ifdef RCFILE
		case 115 /*s*/:
			bit_mask = TKM_SAVE_ON_EXIT;
			#ifdef DEBUG
			   printf("PARSE: SAVE");
			#endif
			break;
		#endif
		default :
			bit_mask=0; 
			break;
			// Do nothing, but it's probably a wrong argv.
			// Any xview command of two char long ?
	   };
	   if (bit_mask !=0) {
		if ( argv[i][0] == 43 ) {		/* + */
		   mix_dat.flags |= bit_mask;
		   #ifdef DEBUG
		     printf(" turned On\n");
		   #endif
		} else {				/* - */
		   mix_dat.flags &= ~bit_mask;
		   #ifdef DEBUG
		     printf(" turned Off\n");
		   #endif
	   	}
	   continue;
	   }
	}
	if (!strcmp(argv[i], "-v1")) {
	   /* View Vol Only */
	   mix_dat.flags = (mix_dat.flags & ~TKM_DISPLAY_WIDTH_MASK) | TKM_DISPLAY_WIDTH_ONE;
	   #ifdef DEBUG
	     printf("PARSE: View One\n");
	   #endif
	   continue;
	}
	if (!strcmp(argv[i], "-vc")) {
	   /* view custom */
	   mix_dat.flags = (mix_dat.flags & ~TKM_DISPLAY_WIDTH_MASK) | TKM_DISPLAY_WIDTH_CUSTOM;
	   #ifdef DEBUG
	     printf("PARSE: View Custom\n");
	   #endif
	   continue;
	}
	if (!strcmp(argv[i], "-vl")) {
	   /* View Large */
	   mix_dat.flags = (mix_dat.flags & ~TKM_DISPLAY_WIDTH_MASK) | TKM_DISPLAY_WIDTH_LARGE;
	   #ifdef DEBUG
	      printf("PARSE: View Large\n");
	   #endif
	   continue;
	}
	if (!strcmp(argv[i], "-vd")) {
	   /* View Debug */
	   mix_dat.flags = (mix_dat.flags & ~TKM_DISPLAY_WIDTH_MASK) | TKM_DISPLAY_WIDTH_ALL;
	   #ifdef DEBUG
	   printf("PARSE: View All devices (Debug)\n");
	   #endif
	   continue;
	}
	if (!strncmp(argv[i], "--custom=",9)) { // ATTENTION strlen("--custom")
	   /* Custom View Order */
	   #ifdef DEBUG
		printf("PARSE: --custom=");
	   #endif
	   for (x = 0 ; x < 20 ; x++) deviceCustomTemp[x] = 1;
	   len = strlen( argv[i] ) ;
	   x = 9;
	   pos =0;
	   device = 0;
	   while ( x < len ) {
	      if ( (argv[i][x] >= 48) & (argv[i][x]<=57) ) {
		//printf("#%d",argv[i][x]);
		device = device * 10 + argv[i][x] - 48;
	      } else {
		device_index_custom[pos] = device;
		deviceCustomTemp[device] = 0;
		pos++;
	        #ifdef DEBUG
		   printf("%d ",device);
	 	#endif
		device =0;
	      }     
	      x++;
	   }
	   if ( device != 0) {
		device_index_custom[pos++] = device;
		deviceCustomTemp[device] = 0;
	        #ifdef DEBUG 
		   printf("%d ",device);
	 	#endif
	   }
	   device_index_custom[pos++] = 0xFF;
	   #ifdef DEBUG
	     printf(" <0xFF> ",device);
	   #endif
	   for (x = 0 ;x <20 ; x++)
		if (deviceCustomTemp[x] != 0) {
		   device_index_custom[pos++] = x;
		   #ifdef DEBUG
		      printf("%d ",x);
		   #endif
		}
	   #ifdef DEBUG
	     printf("\n");
	   #endif
	   continue;
	}
	else {
	#ifdef DEBUG
	  printf("!=CUSTOM: %s",argv[i] );
	#endif
	}
//	// hey I can't do this.. to support xview command lines !
//	fprintf(stderr,"Unknown option: %s\n",argv[i]);
//	usage();
//	exit(1);
        }
}




int main( int argc , char *argv[])
{
int i;
for (i = 1 ; i<=20 ;i++)  
  device_index_custom[i]= i;
  device_index_custom[21] = 0xFF;

   mix_dat.flags=TKM_DISPLAY_DEFAULTVALUE;
   strcpy(mixer_device, TKM_MIXER_DEVICE);
   #ifdef RCFILE
   ReadRcFile();
   #endif // ifdef RCFILE
  if (!parse_args(argc,argv)) {
      usage();
      exit(1);
  }
  if (!GetMixerInfo()) {
      exit(1);
  }

  quit(); 
  xv_init(
//	FRAME_SCALE_STATE,FRAME_SCALE_LARGE,
	XV_INIT_ARGC_PTR_ARGV, &argc, argv,
	NULL );
  BuildXViewInterface();
  RefreshXViewInterface();
  #ifdef PREFS
  BuildXViewPrefFrame();
  RefreshXViewPrefFrame();
  #endif // if PREFS
  if(mix_dat.flags & TKM_POLL)  adjust_timer_speed(TKM_POLL_DEFAULT_INTERVAL);
  update_all();
  xv_main_loop(frame);
  #ifdef RCFILE
  if ((mix_dat.flags & TKM_SAVE_ON_EXIT)!=0) SaveRcFile();
  #endif // ifdef RCFILE
  exit(0);
}

